/*
 * libdevice-node
 *
 * Copyright (c) 2018 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <hw/usb_gadget.h>
#include <hw/board.h>

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>

#include <libsyscommon/dbus-systemd.h>

#define zalloc(amount) calloc(1, amount)

/* Based on slp-gadget and initial version of USB HAL by Taeyoung Kim */
#define DEFAULT_VID 0x04e8
#define DEFAULT_PID 0x6860
#define DEFAULT_BCD_DEVICE 0x0100

#define DEFAULT_LANG 0x409 /* US_en */
#define DEFAULT_MANUFACTURER "Samsung"
#define DEFAULT_PRODUCT "TIZEN"
#define DEFAULT_SERIAL "01234TEST"

#define DEFAULT_BMATTRIBUTES ((1 << 7) | (1 << 6))
#define DEFAULT_MAX_POWER 500

#ifndef EXPORT
#define EXPORT	__attribute__ ((visibility("default")))
#endif

static void simple_cleanup_config(struct usb_configuration *config)
{
	int i;

	if (!config)
		return;

	if (config->strs) {
		for (i = 0; config->strs[i].lang_code; ++i)
			free(config->strs[i].config_str);

		free(config->strs);
	}

	/*
	 * Each function will be free later,
	 * for now we cleanup only pointers.
	 */
	if (config->funcs)
		free(config->funcs);

	free(config);
}

static void simple_cleanup_gadget(struct usb_gadget *gadget)
{
	int i;

	if (!gadget)
		return;

	if (gadget->strs) {
		for (i = 0; gadget->strs[i].lang_code; ++i) {
			free(gadget->strs[i].manufacturer);
			free(gadget->strs[i].product);
			free(gadget->strs[i].serial);
		}
		free(gadget->strs);
	}

	if (gadget->configs) {
		for (i = 0; gadget->configs[i]; ++i)
			simple_cleanup_config(gadget->configs[i]);

		free(gadget->configs);
	}

	if (gadget->funcs) {
		for (i = 0; gadget->funcs[i]; ++i)
			gadget->funcs[i]->free_func(gadget->funcs[i]);

		free(gadget->funcs);
	}

	free(gadget);
}

static int alloc_default_config(struct usb_configuration **_config)
{
	struct usb_configuration *config;

	config = zalloc(sizeof(*config));
	if (!config)
		goto out;

	config->strs = calloc(1, sizeof(*config->strs));
	if (!config->strs)
		goto free_config;

	config->attrs.bmAttributs = DEFAULT_BMATTRIBUTES;
	config->attrs.MaxPower = DEFAULT_MAX_POWER;

	*_config = config;

	return 0;

free_config:
	free(config);
out:
	return -ENOMEM;
}

static int get_device_serial(char **out)
{
	struct hw_info *info;
	struct hw_common *common;
	struct hw_board *board;
	int ret;

	if (hw_get_info(BOARD_HARDWARE_DEVICE_ID, (const struct hw_info **)&info))
		return -ENOENT;

	ret = info->open(info, NULL, &common);
	if (ret < 0)
		return ret;

	board = container_of(common, struct hw_board, common);
	return board->get_device_serial(out);
}

static int alloc_default_gadget(struct usb_gadget **_gadget)
{
	struct usb_gadget *gadget;
	struct usb_gadget_strings *strs;
	struct usb_configuration **configs;
	int ret;

	gadget = zalloc(sizeof(*gadget));
	if (!gadget)
		goto out;

	gadget->attrs.idVendor = DEFAULT_VID;
	gadget->attrs.idProduct = DEFAULT_PID;
	gadget->attrs.bcdDevice = DEFAULT_BCD_DEVICE;

	strs = calloc(2, sizeof(*strs));
	if (!strs)
		goto free_gadget;

	strs[0].lang_code = 0x409;
	strs[0].manufacturer = strdup(DEFAULT_MANUFACTURER);
	strs[0].product = strdup(DEFAULT_PRODUCT);
	ret = get_device_serial(&strs[0].serial);
	if (ret < 0)
		strs[0].serial = strdup(DEFAULT_SERIAL);

	if (!strs[0].manufacturer || !strs[0].product || !strs[0].serial)
		goto free_strs;

	gadget->strs = strs;

	/* slp-gadget use max 2 confiuration and NULL termination */
	configs = calloc(3, sizeof(*configs));
	if (!configs)
		goto free_strs;

	gadget->configs = configs;
	*_gadget = gadget;

	return 0;

free_strs:
	free(strs[0].manufacturer);
	free(strs[0].product);
	free(strs[0].serial);
	free(strs);
free_gadget:
	free(gadget);
out:
	return -ENOMEM;
}

static inline struct usb_function *find_func(struct usb_gadget *gadget,
					    int func_id)
{
	int i;

	for (i = 0; gadget->funcs[i] && gadget->funcs[i]->id != func_id; ++i);

	return gadget->funcs[i];
}

static int simple_id_to_gadget(struct usb_gadget_id *gadget_id,
			       struct usb_gadget **_gadget)
{
	struct usb_gadget *gadget;
	int n_configs = 0;
	/* zero terminates */
	int functions[2][sizeof(gadget_id->function_mask)*8];
	int n_functions;
	struct usb_function **funcs;
	int idx, i, j;
	int ret;

	if (!gadget_id || !_gadget)
		return -EINVAL;

	ret = alloc_default_gadget(&gadget);
	if (ret)
		goto out;

	/*
	 * Currently all gadgets use inly single configuration but
	 * slp-gadget is capable to handle two of them
	 *
	 * Order of interfaces in configuration is significant
	 * so in this switch we sort our functions in a correct order
	 */
	switch (gadget_id->function_mask) {
	case USB_FUNCTION_MTP:
		n_configs = 1;
		functions[0][0] = USB_FUNCTION_MTP;
		functions[0][1] = 0;
		gadget->attrs.idProduct = 0x6860;
		break;

	case USB_FUNCTION_ACM:
		n_configs = 1;
		functions[0][0] = USB_FUNCTION_ACM;
		functions[0][1] = 0;
		gadget->attrs.idProduct = 0x6860;
		break;

	case USB_FUNCTION_SDB:
		n_configs = 1;
		functions[0][0] = USB_FUNCTION_SDB;
		functions[0][1] = 0;
		gadget->attrs.idProduct = 0x685d;
		break;

	/* MTP, ACM, SDB */
	case USB_FUNCTION_MTP | USB_FUNCTION_ACM:
		n_configs = 1;
		functions[0][0] = USB_FUNCTION_MTP;
		functions[0][1] = USB_FUNCTION_ACM;
		functions[0][2] = 0;
		gadget->attrs.idProduct = 0x6860;
		break;

	case USB_FUNCTION_ACM | USB_FUNCTION_SDB:
		n_configs = 1;
		functions[0][0] = USB_FUNCTION_ACM;
		functions[0][1] = USB_FUNCTION_SDB;
		functions[0][2] = 0;
		gadget->attrs.idProduct = 0x6860;
		break;

	case USB_FUNCTION_MTP | USB_FUNCTION_SDB:
		n_configs = 1;
		functions[0][0] = USB_FUNCTION_MTP;
		functions[0][1] = USB_FUNCTION_SDB;
		functions[0][2] = 0;
		gadget->attrs.idProduct = 0x6860;
		break;

	case USB_FUNCTION_MTP | USB_FUNCTION_ACM | USB_FUNCTION_SDB:
		n_configs = 1;
		functions[0][0] = USB_FUNCTION_MTP;
		functions[0][1] = USB_FUNCTION_ACM;
		functions[0][2] = USB_FUNCTION_SDB;
		functions[0][3] = 0;
		gadget->attrs.idProduct = 0x6860;
		break;

	/* DIAG */
	case USB_FUNCTION_MTP | USB_FUNCTION_ACM | USB_FUNCTION_SDB | USB_FUNCTION_DIAG:
		n_configs = 1;
		functions[0][0] = USB_FUNCTION_MTP;
		functions[0][1] = USB_FUNCTION_ACM;
		functions[0][2] = USB_FUNCTION_SDB;
		functions[0][3] = USB_FUNCTION_DIAG;
		functions[0][4] = 0;
		gadget->attrs.idProduct = 0x6860;
		break;

	/* RNDIS */
	case USB_FUNCTION_RNDIS:
		n_configs = 1;
		functions[0][0] = USB_FUNCTION_RNDIS;
		functions[0][1] = 0;
		gadget->attrs.idProduct = 0x6863;
		break;

	case USB_FUNCTION_SDB | USB_FUNCTION_RNDIS:
		n_configs = 1;
		functions[0][0] = USB_FUNCTION_SDB;
		functions[0][1] = USB_FUNCTION_RNDIS;
		functions[0][2] = 0;
		gadget->attrs.idProduct = 0x6864;
		break;

	case USB_FUNCTION_RNDIS | USB_FUNCTION_DIAG:
		n_configs = 1;
		functions[0][0] = USB_FUNCTION_RNDIS;
		functions[0][1] = USB_FUNCTION_DIAG;
		functions[0][2] = 0;
		gadget->attrs.idProduct = 0x6864;
		break;

	case USB_FUNCTION_ACM | USB_FUNCTION_SDB | USB_FUNCTION_RNDIS:
		n_configs = 1;
		functions[0][0] = USB_FUNCTION_ACM;
		functions[0][1] = USB_FUNCTION_SDB;
		functions[0][2] = USB_FUNCTION_RNDIS;
		functions[0][3] = 0;
		gadget->attrs.idProduct = 0x6864;
		break;

	/* RMNET */
	case USB_FUNCTION_DIAG | USB_FUNCTION_RMNET:
		n_configs = 1;
		functions[0][0] = USB_FUNCTION_DIAG;
		functions[0][1] = USB_FUNCTION_RMNET;
		functions[0][2] = 0;
		gadget->attrs.idProduct = 0x685d;
		break;

	/* DM */
	case USB_FUNCTION_ACM | USB_FUNCTION_SDB | USB_FUNCTION_DM:
		n_configs = 1;
		functions[0][0] = USB_FUNCTION_ACM;
		functions[0][1] = USB_FUNCTION_SDB;
		functions[0][2] = USB_FUNCTION_DM;
		functions[0][3] = 0;
		gadget->attrs.idProduct = 0x6860;
		break;

	default:
		ret = -EINVAL;
		goto free_gadget;
	};

	n_functions = __builtin_popcount(gadget_id->function_mask);

	funcs = calloc(n_functions + 1, sizeof(*funcs));
	if (!funcs) {
		ret = -ENOMEM;
		goto free_gadget;
	}

	gadget->funcs = funcs;

	idx = 0;
	for (i = 0; _available_funcs[i]; ++i) {
		int func_id = 1 << i;

		if (!(gadget_id->function_mask & func_id))
			continue;

		ret = _available_funcs[i]->clone(_available_funcs[i],
			gadget->funcs + idx);
		if (ret)
			goto free_functions;
		++idx;
	}

	for (j = 0; j < n_configs; ++j) {
		struct usb_configuration *config;
		int n_funcs_in_config;

		for (i = 0; functions[j][i]; ++i);
		n_funcs_in_config = i;

		ret = alloc_default_config(&config);
		if (ret)
			goto free_configs;

		gadget->configs[j] = config;
		config->funcs = calloc(n_funcs_in_config + 1,
						       sizeof(void *));
		if (!config->funcs)
			goto free_configs;

		for (i = 0; functions[j][i]; ++i)
			config->funcs[i] = find_func(gadget, functions[j][i]);
	}

	*_gadget = gadget;
	return 0;
free_configs:
free_functions:
free_gadget:
	simple_cleanup_gadget(gadget);
out:
	return ret;
}

void rndis_handler(int enable)
{
	if (enable)
		(void)systemd_start_unit_wait_started("rndis.service", NULL, -1);
	else
		(void)systemd_stop_unit_wait_stopped("rndis.service", NULL, -1);
}

static void free_simple_func(struct usb_function *func)
{
	if (func) {
		free((void *)func->name);
		free((void *)func->instance);
		free((void *)func->ffs_service);
		free((void *)func->service);
		free(func);
	}
}

static int clone_simple_func(struct usb_function *func,
			      struct usb_function **clone)
{
	struct usb_function *other;

	if (!func || !clone)
		return -EINVAL;

	other = (struct usb_function *)calloc(1, sizeof(struct usb_function));
	if (!other)
		return -ENOMEM;

	*other = *func;

	other->name = strdup(func->name);
	other->instance = strdup(func->instance);
	if (!other->name || !other->instance)
		goto out_nomem;

	if (func->ffs_service) {
		other->ffs_service = strdup(func->ffs_service);
		if (!other->ffs_service)
			goto out_nomem;
	}

	if (func->service) {
		other->service = strdup(func->service);
		if (!other->service)
			goto out_nomem;
	}

	*clone = other;
	return 0;

out_nomem:
	free_simple_func(other);
	return -ENOMEM;
}

#define DEFINE_USB_FUNCTION(_id, _name, _ffs_service, _service, _handler)	\
	static struct usb_function _##_name##_function = {                      \
		.id = _id,                                                      \
		.name = #_name,                                                 \
		.instance = "default",                                          \
		.ffs_service = _ffs_service,                                    \
		.service = _service,                                            \
		.handler = _handler,                                            \
		.clone = clone_simple_func,                                     \
		.free_func = free_simple_func,                                  \
	}

DEFINE_USB_FUNCTION(USB_FUNCTION_DIAG,        diag,        NULL, NULL,            NULL);
DEFINE_USB_FUNCTION(USB_FUNCTION_RMNET,       rmnet,       NULL, NULL,            NULL);
DEFINE_USB_FUNCTION(USB_FUNCTION_DM,          dm,          NULL, NULL,            NULL);
DEFINE_USB_FUNCTION(USB_FUNCTION_CONN_GADGET, conn_gadget, NULL, NULL,            NULL);
DEFINE_USB_FUNCTION(USB_FUNCTION_SDB,         sdb,         "sdbd", NULL,          NULL);
DEFINE_USB_FUNCTION(USB_FUNCTION_MTP,         mtp,         "mtp-responder", NULL, NULL);
DEFINE_USB_FUNCTION(USB_FUNCTION_ACM,         acm,         NULL, "data-router",   NULL);
DEFINE_USB_FUNCTION(USB_FUNCTION_RNDIS,       rndis,       NULL, "sshd",          rndis_handler);
#undef DEFINE_USB_FUNCTION

/* Caution: index order of arrary is important, because simple_id_to_gadget() uses it. */
struct usb_function *_available_funcs[] = {
	[USB_FUNCTION_IDX_MTP]         = &_mtp_function,
	[USB_FUNCTION_IDX_ACM]         = &_acm_function,
	[USB_FUNCTION_IDX_SDB]         = &_sdb_function,
	[USB_FUNCTION_IDX_RNDIS]       = &_rndis_function,
	[USB_FUNCTION_IDX_DIAG]        = &_diag_function,
	[USB_FUNCTION_IDX_CONN_GADGET] = &_conn_gadget_function,
	[USB_FUNCTION_IDX_DM]          = &_dm_function,
	[USB_FUNCTION_IDX_RMNET]       = &_rmnet_function,
	[USB_FUNCTION_IDX_MAX]         = NULL /* An indicator to end the array */
};

EXPORT
int simple_translator_open(struct hw_info *info,
		const char *id, struct hw_common **common)
{
	struct usb_gadget_translator *simple_translator;

	if (!info || !common)
		return -EINVAL;

	simple_translator = zalloc(sizeof(*simple_translator));
	if (!simple_translator)
		return -ENOMEM;

	simple_translator->common.info = info;
	simple_translator->id_to_gadget = simple_id_to_gadget;
	simple_translator->cleanup_gadget = simple_cleanup_gadget;

	/* Use mtp-responder-dummy.socket when there is no mtp-responser.socket.
	 *
	 * The mtp-responder.socket is special in the configfs environment.
	 * If mtp-responder.socket is missing, gadget configuration will fail.
	 * As a result, all usb operations do not work properly.
	 * So in environments that mtp doesn't support, use dummy mtp.
	 */
	if (access("/usr/lib/systemd/system/mtp-responder.socket", F_OK)) {
		_available_funcs[USB_FUNCTION_IDX_MTP]->ffs_service = "mtp-responder-dummy";
	}

	*common = &simple_translator->common;
	return 0;
}

EXPORT
int simple_translator_close(struct hw_common *common)
{
	struct usb_gadget_translator *simple_translator;

	if (!common)
		return -EINVAL;

	simple_translator = container_of(common, struct usb_gadget_translator,
					 common);

	free(simple_translator);
	return 0;
}
